/**********************************************************************
 * pop3_token.c                                              March 1999
 * Horms                                            horms@verge.ing.to.
 * Simon Horman                               http://www.verge.ing.to./
 *
 * Toy pop3 server
 **********************************************************************/

#include "pop3_token.h"


/**********************************************************************
 * read_pop3_line
 * read a line from fd and parse it into a queue of pop3_tokens
 **********************************************************************/

queue *read_pop3_line(int fd, queue *q){
  pop3_token *token;

  if((q=create_queue(q))==NULL){
    syslog(LOG_WARNING, "read_pop3_line: create_queue");
    return(NULL);
  }
 
  do{
    if((token=read_pop3_token(fd, token))==NULL){
      syslog(LOG_WARNING, "read_pop3_line: read_pop3_token");
      return(NULL);
    }

    if((q=push(q, (queue_t)token, DESTROY_POP3_TOKEN))==NULL){
      syslog(LOG_WARNING, "read_pop3_line: push");
      return(NULL);
    }
  }while(!eol_pop3_token(token));

  return(q);
}


/**********************************************************************
 * create_pop3_token
 * create an empty token
 *
 * 8 bit clean
 **********************************************************************/

pop3_token *create_pop3_token(pop3_token *token){
  if((token=(pop3_token *)malloc(sizeof(pop3_token)))==NULL){
    syslog(LOG_WARNING, "create_pop3_token: malloc");
    return(NULL);
  }
  token->n=0;
  token->buf=NULL;
  return(token);
}


/**********************************************************************
 * assign_pop3_token
 * place bytes into a token
 *
 * 8 bit clean
 **********************************************************************/

void assign_pop3_token(
  pop3_token *token, 
  unsigned char * buf, 
  ssize_t n,
  int eol
){
  token->n=n;
  token->eol=eol;
  if(n==0){
    token->buf=NULL;
    free(buf);
  }
  else {
    token->buf=buf;
  }
}


/**********************************************************************
 * unassign_pop3_token
 * make a token empty
 *
 * useful if you want to destroy a token but not what it contains
 *
 * 8 bit clean
 **********************************************************************/

void unassign_pop3_token(pop3_token *token){
  assign_pop3_token(token, (unsigned char *)NULL, (ssize_t)0, 0);
}

/**********************************************************************
 * destroy_pop3_token
 * if the token is null no nothing
 * if buf is non-NULL then free it
 * free the token
 *
 * 8 bit clean
 **********************************************************************/

void destroy_pop3_token(pop3_token *token){
  if(token==NULL){
    return;
  }

  if(token->buf!=NULL){
    free(token->buf);
  }

  free(token);
}


/**********************************************************************
 * write_pop3_token
 * write a token to fd
 *
 * 8 bit clean
 **********************************************************************/

int write_pop3_token(int fd, pop3_token *token){
  ssize_t offset;
  ssize_t bytes_written;

  offset=0;
  do{
    bytes_written=write(fd, (token->buf)+offset, (token->n)-offset);
    if(bytes_written<0){
      syslog(
        LOG_WARNING, 
        "write_pop3_token: error writing output: %s",
        strerror(errno)
      );
      return(-1);
    }
    offset+=bytes_written;
  }while(offset<token->n);
  
  return(0);
}


/**********************************************************************
 * write_strings
 * write strings to fd by puting them into pop3_tokens and
 * printing the tokens
 * Extreemely inefficient but does use the same write methods
 * as tokens
 * 
 * Not 8 bit clean
 **********************************************************************/

int write_strings(int fd, int nostring, ...){
  pop3_token *token;
  va_list ap;
  char *string;

  if(nostring<1){
    return(1);
  }

  if((token=create_pop3_token(token))==NULL){
    syslog(LOG_WARNING, "write_strings: create_pop3_token");
    return(-1);
  }

  va_start(ap, nostring);
  while(nostring-->0){
    string=va_arg(ap, char*);
    if(string==NULL){
      syslog(LOG_WARNING, "write_strings: null string");
      return(-1);
    }
    assign_pop3_token(token, (unsigned char *)string, strlen(string), -1);
    if(write_pop3_token(fd, token)<0){
      syslog(LOG_WARNING, "write_strings: write_pop3_token");
      return(-1);
    }
  }
  va_end(ap);

  unassign_pop3_token(token);
  destroy_pop3_token(token);

  return(0);
}


/**********************************************************************
 * read_pop3_token
 * read a token in from fd
 * ' ' will terminate a token
 * '\r' is ignored
 * '\n' will terminate a token and set the eol element to 1
 * all other characters are considered to be a part of the token
 *
 * 8 bit clean
 **********************************************************************/

pop3_token *read_pop3_token(int fd, pop3_token *token){
  unsigned char *buffer;
  int bytes_read;
  int offset;
  int eol;

  if(
    (buffer=(unsigned char *)malloc(sizeof(unsigned char)*BUFFER_SIZE))==NULL
  ){
    syslog(
      LOG_WARNING, 
      "read_pop3_token: malloc buffer: %s", 
      strerror(errno)
    );
    return(NULL);
  }

  offset=0;
  eol=0;
  while(offset<BUFFER_SIZE){
    if((bytes_read=read(fd, buffer+offset, 1))<0){
      syslog(
        LOG_WARNING, 
        "read_pop3_token: error reading input",
        strerror(errno)
      );
      return(NULL);
    }
    if(bytes_read==0){
      continue;
    }

    if(*(buffer+offset)=='\n'){
      eol=1;
      break;
    }
    else if(*(buffer+offset)=='\r'){
      continue;
    }
    else if(*(buffer+offset)==' '){
      break;
    }
    else{
      offset++;
    }
  }

  if((token=create_pop3_token(token))==NULL){
    syslog(LOG_WARNING, "read_pop3_token: create_pop3_token");
    return(NULL);
  }

  assign_pop3_token(token, buffer, offset, eol);
    
  return(token);
}

int eol_pop3_token(pop3_token *token){
  return(token->eol);
}


/**********************************************************************
 * pop3_token_cmp
 * compare two tokens
 * return 0 is they are the same, -1 otherwise
 * eol field will be ignored if eiter token has eol 
 * set to -1 (don't care)
 *
 * Not 8 bit clean as it is case insensitive using toupper
 **********************************************************************/

int pop3_token_cmp(pop3_token *a, pop3_token *b){
  int i;

  if(a->n!=b->n){
    return(-1);
  }

  if((a->eol!=b->eol) && (a->eol>-1) && (b->eol>-1)){
    return(-1);
  }

  for(i=0;i<a->n;i++){
    if(toupper(*(a->buf+i))!=toupper(*(b->buf+i))){
      return(-1);
    }
  }

  return(0);
}


/**********************************************************************
 * pop3_token_to_string
 * dump the buffer in a pop3_token into a \0 terminated string
 * string will be dynamically allocated
 *
 * Sting assigned is 8 bit clean
 **********************************************************************/

char *pop3_token_to_string(pop3_token *token, char *string){
  if((string=strn_to_str(string, (char *)token->buf, token->n))==NULL){
    syslog(LOG_WARNING, "pop3_token_to_string: strn_to_str");
    return(NULL);
  }
  return(string);
}

